# -*- coding: utf-8 -*-
########################################################################
# 
# Copyright (c) 2019 Baidu.com, Inc. All Rights Reserved
# 
########################################################################
 
from .gen.cluster import Cluster
from .gen.algorithm import generate_regions
from .gen.algorithm import rect_cluster
from .gen.algorithm import rect_classify
from .gen.algorithm import region_filter
from .gen.algorithm import split_segs_by_segs
from .gen.algorithm import splitedsegs_2_regions
from .gen.algorithm import merge_regions
from .gen.regionalg import RegionFinder

